import json
from swap.common.schemas import ServiceDiscoveryResponseSchema
from swap.common.filters import ServiceDiscoveryResponseFilter


class Parser:
    def apply_pre_processing(self, response):
        return json.loads(response)

    def apply_filters(self, response):
        return self.filter.apply(response)

    def parse(self, response):
        preprocessed_response = self.apply_pre_processing(response)
        filtered_response = self.apply_filters(preprocessed_response)

        return self.schema.load(filtered_response)


class ServiceDiscoveryResponseParser(Parser):
    def __init__(self):
        self.filter = ServiceDiscoveryResponseFilter()
        self.schema = ServiceDiscoveryResponseSchema()



from swap.common.logging import logger


class ServiceRequestIdentity:
    def match(self, service_request, config):
        service_catalogue_uuid = service_request.service_catalogue_uuid
        service_name = config.name

        if service_catalogue_uuid != service_name:
            logger.info(f'ServiceRequest service_catalogue_uuid did not match, ignoring message')
            logger.info(
                f'Message for {service_catalogue_uuid} not {service_name}'
            )

            return False

        return True


class OfferExecutionRequestIdentity:
    def match(self, offer_execution_request, config):
        service_offer_name = offer_execution_request.service_offer_name
        service_name = config.name

        if service_offer_name != service_name:
            logger.info(f'OfferExecutionRequest service_offer_name did not match, ignoring message')
            logger.info(
                f'Message for {service_offer_name} not {service_name}'
            )

            return False

        service_provider_header = offer_execution_request.service_provider_header

        attributes = [
            'service_provider_uuid',
            'service_provider_signature'
        ]

        for attribute in attributes:
            offer_execution_request_attribute = getattr(service_provider_header, attribute)
            config_attribute = getattr(config, attribute)

            if offer_execution_request_attribute != config_attribute:
                logger.info(f'OfferExecutionRequest {attribute} did not match, ignoring message')
                logger.info(
                    f'Message for {offer_execution_request_attribute} not {config_attribute}'
                )

                return False

        return True



from swap.api.clients import UBLClient
from swap.api.parsers import InputDatasetsResponseParser
from swap.common.config import Settings
import json

client = UBLClient()

class Dataset:
    def __init__(self, input_dataset, metadata=None):
        self.name = input_dataset.name
        self.input_dataset = input_dataset
        self.dataset_type = 'GCS'
        self.url = f'gs://{Settings.BUCKET}/{input_dataset.name}'
        self.metadata = metadata

    def to_dict(self):
        dataset_dict = {}
        dataset_dict['data_uri'] = self.url
        dataset_dict['metadata'] = json.dumps(self.metadata)
        dataset_dict['metadata'] = dataset_dict['metadata'].replace('"', "'")
        return dataset_dict


class CDFDataset:
    def __init__(self, dataset_id, survey_id="N/A", metadata=None):
        if survey_id != "N/A":
            self.url = f'cdf://{survey_id}/{dataset_id}'
        else:
            self.url = f'cdf://{dataset_id}'
        self.name = dataset_id
        self.survey_id = survey_id
        self.dataset_id = dataset_id
        self.dataset_type = 'CDF'
        self.metadata = metadata

    def to_dict(self):
        dataset_dict = {}
        dataset_dict['data_uri'] = self.url
        dataset_dict['metadata'] = json.dumps(self.metadata)
        dataset_dict['metadata'] = dataset_dict['metadata'].replace('"', "'")
        return dataset_dict


def list_all():
    raw_result = client.get_input_datasets()
    parser = InputDatasetsResponseParser()
    input_datasets_response = parser.parse(raw_result).items
    output = []

    for input_dataset in input_datasets_response:
        dataset = Dataset(input_dataset=input_dataset)
        output.append(dataset)

    return output


def select(name):
    datasets = list_all()

    for dataset in datasets:
        if dataset.name == name:
            return dataset

    return None



from swap.messaging.actions import (
    DefaultMessageAction,
    ExecutionRequestCancelAction,
    IgnoredMessageAction,
    OfferExecutionRequestAction,
    ServiceDiscoveryAction,
    ServiceRequestAction,
)
from swap.messaging.callbacks import DefaultCallback
from swap.messaging.exceptions import InvalidActionException


class Registry:
    def __init__(self, config):
        self.config = config
        self.actions = {}
        self.callbacks = {}

    def _validate(self, action):
        call_method = getattr(action, 'call', None)

        if not call_method:
            raise InvalidActionException('Action missing call attribute')

        if not callable(call_method):
            raise InvalidActionException('Action call attribute not callable')

    def register_action(self, message_type, action):
        self._validate(action)
        self.actions[message_type] = action

    def register_callback(self, message_type, callback):
        self.callbacks[message_type] = callback

    def action(self, message_type):
        return self.actions.get(message_type, DefaultMessageAction(config=self.config))

    def callback(self, message_type):
        return self.callbacks.get(message_type, DefaultCallback)


class DefaultRegistry(Registry):
    def __init__(self, config):
        self.config = config

        self.actions = {
            'ServiceDiscovery': ServiceDiscoveryAction(config=config),
            'ServiceDiscoveryResponse': IgnoredMessageAction(config=config),
            'ServiceRequest': ServiceRequestAction(config=config),
            'ServiceOffering': IgnoredMessageAction(config=config),
            'ExecutionRequest': IgnoredMessageAction(config=config),
            'ExecutionRequestCancel': ExecutionRequestCancelAction(config=config),
            'ExecutionRequestAccept': IgnoredMessageAction(config=config),
            'ExecutionRequestDecline': IgnoredMessageAction(config=config),
            'ExecutionStatusUpdate': IgnoredMessageAction(config=config),
            'ExecutionCompletion': IgnoredMessageAction(config=config),
            'ExecutionFailure': IgnoredMessageAction(config=config),
            'OfferExecutionRequest': OfferExecutionRequestAction(config=config),
        }

        self.callbacks = {}



import json
import time
from uuid import uuid4
from swap.common.models import (
    ServiceDiscoveryData,
    ServiceDiscoveryDataset,
    ServiceDiscoveryIcon,
    ServiceDiscoveryParameters,
    ServiceDiscoveryResponse,
    ServiceDiscoveryService,
    ServiceDiscoveryVendor,
)
from swap.common.schemas import (
    ServiceDiscoveryResponseSchema
)
from swap.messaging.models import (
    ServiceOffering,
    ServiceProviderHeader,
    ExecutionFailure,
    ExecutionRequestAccept,
    ExecutionStatusUpdate,
    ExecutionCompletion,
)
from swap.messaging.schemas import (
    ServiceOfferingSchema,
    ExecutionFailureSchema,
    ExecutionRequestAcceptSchema,
    ExecutionStatusUpdateSchema,
    ExecutionCompletionSchema,
)


class ServiceDiscoveryResponseBuilder:
    def __init__(self, config):
        self.config = config

    def service_discovery_vendor(self):
        return ServiceDiscoveryVendor(
            service_provider_name=self.config.service_provider_name,
            service_provider_uuid=self.config.service_provider_uuid,
            service_provider_signature=self.config.service_provider_signature
        )

    def service_discovery_parameters(self):
        return ServiceDiscoveryParameters(
            properties=self.config.parameters
        )

    def service_discovery_service(self):
        return ServiceDiscoveryService(
            name=self.config.name,
            catalogue_uuid=self.config.catalogue_uuid,
            service_description=self.config.service_description,
            parameters=self.service_discovery_parameters()
        )

    def service_discovery_dataset(self):
        return ServiceDiscoveryDataset(
            input=self.config.dataset_type,
            output=self.config.dataset_type
        )

    def service_discovery_icon(self):
        return ServiceDiscoveryIcon(
            base64=self.config.icon_base64,
            url=self.config.icon_url
        )

    def service_discovery_response_data(self):
        return ServiceDiscoveryData(
            vendor=self.service_discovery_vendor(),
            service=self.service_discovery_service(),
            dataset=self.service_discovery_dataset(),
            icon=self.service_discovery_icon()
        )

    def service_discovery_response(self):
        return ServiceDiscoveryResponse(
            data=[self.service_discovery_response_data()]
        )

    def build(self):
        schema = ServiceDiscoveryResponseSchema()
        output = schema.dump(self.service_discovery_response())

        return json.dumps(output, indent=2)


class ServiceOfferingBuilder:
    def __init__(self, message_id, config, service_request):
        self.config = config
        self.service_request = service_request
        self.message_id = message_id

    def service_provider_header(self):
        return ServiceProviderHeader(
            service_provider_uuid=self.config.service_provider_uuid,
            service_offer_uuid=str(uuid4()),
            service_request_uuid=self.message_id,
            service_provider_signature=self.config.service_provider_signature,
        )

    def build(self):
        self.service_request.standard_header.request_message_uuid = self.message_id
        self.service_request.standard_header.originating_message_uuid = self.message_id

        service_offering = ServiceOffering(
            standard_header=self.service_request.standard_header,
            service_provider_header=self.service_provider_header(),
            input_data=self.service_request.input_data,
            output_data=self.service_request.output_data,
            service_offer_description=self.config.service_description,
            service_offer_parameters=self.config.parameters,
            service_offer_name=self.config.name,
        )

        schema = ServiceOfferingSchema()
        output = schema.dump(service_offering)

        return json.dumps(output, indent=2)


class ExecutionFailureBuilder:
    def __init__(
        self, message_id, offer_execution_request, execution_details={}, exec_fail_code='',
        exec_fail_text=''
    ):
        self.message_id = message_id
        self.offer_execution_request = offer_execution_request
        self.execution_details = execution_details
        self.exec_fail_code = exec_fail_code
        self.exec_fail_text = exec_fail_text

    def build(self):
        unfulfilled_uuid = self.offer_execution_request.service_provider_header.service_offer_uuid

        self.offer_execution_request.standard_header.request_message_uuid = self.message_id

        execution_failure = ExecutionFailure(
            standard_header=self.offer_execution_request.standard_header,
            service_provider_header=self.offer_execution_request.service_provider_header,
            execution_details=self.execution_details,
            exec_fail_code=self.exec_fail_code,
            exec_fail_text=self.exec_fail_text,
            unfulfilled_service_uuid=unfulfilled_uuid,
        )

        schema = ExecutionFailureSchema()
        output = schema.dump(execution_failure)

        return json.dumps(output, indent=2)


class ExecutionRequestAcceptBuilder:
    def __init__(self, message_id, offer_execution_request, execution_details={}):
        self.message_id = message_id
        self.offer_execution_request = offer_execution_request
        self.execution_details = execution_details

    def build(self):
        self.offer_execution_request.standard_header.request_message_uuid = self.message_id

        execution_request_accept = ExecutionRequestAccept(
            standard_header=self.offer_execution_request.standard_header,
            service_provider_header=self.offer_execution_request.service_provider_header,
            execution_details=self.execution_details,
            exec_start_timestamp=time.time()
        )

        schema = ExecutionRequestAcceptSchema()
        output = schema.dump(execution_request_accept)

        return json.dumps(output, indent=2)


class ExecutionStatusUpdateBuilder:
    def __init__(
        self, message_id, offer_execution_request, execution_details={}, exec_status_code='',
        exec_status_text=''
    ):
        self.message_id = message_id
        self.offer_execution_request = offer_execution_request
        self.execution_details = execution_details
        self.exec_status_code = exec_status_code
        self.exec_status_text = exec_status_text

    def build(self):
        self.offer_execution_request.standard_header.request_message_uuid = self.message_id

        update = ExecutionStatusUpdate(
            standard_header=self.offer_execution_request.standard_header,
            service_provider_header=self.offer_execution_request.service_provider_header,
            execution_details=self.execution_details,
            exec_status_code=self.exec_status_code,
            exec_status_text=self.exec_status_text,
            exec_update_timestamp=time.time()
        )

        schema = ExecutionStatusUpdateSchema()
        output = schema.dump(update)

        return json.dumps(output, indent=2)


class ExecutionCompletionBuilder:
    def __init__(
        self, message_id, offer_execution_request, execution_details={},
        output_data_id = '', exec_comp_code='', exec_comp_text=''
    ):
        self.message_id = message_id
        self.offer_execution_request = offer_execution_request
        self.execution_details = execution_details
        self.output_data_id = output_data_id
        self.exec_comp_code = exec_comp_code
        self.exec_comp_text = exec_comp_text

    def build(self):
        self.offer_execution_request.standard_header.request_message_uuid = self.message_id

        # Add provided output_data_id to results dataset
        results_dataset = self.offer_execution_request.output_data
        results_dataset.file_uri = self.output_data_id

        update = ExecutionCompletion(
            standard_header=self.offer_execution_request.standard_header,
            service_provider_header=self.offer_execution_request.service_provider_header,
            execution_details=self.execution_details,
            exec_comp_code=self.exec_comp_code,
            exec_comp_text=self.exec_comp_text,
            results_dataset=results_dataset,
            exec_completion_timestamp=time.time()
        )

        schema = ExecutionCompletionSchema()
        output = schema.dump(update)

        return json.dumps(output, indent=2)



from swap.api.clients import ServiceDiscoveryClient
from swap.api.models import Service
from swap.common.parsers import ServiceDiscoveryResponseParser


client = ServiceDiscoveryClient()


def list_all():
    raw_result = client.get_services()
    parser = ServiceDiscoveryResponseParser()
    service_discovery_response = parser.parse(raw_result).data
    output = []

    for service_discovery_data in service_discovery_response:
        service = Service(service_discovery_data)
        output.append(service)

    return output


def select(name):
    services = list_all()

    for service in services:
        if service.name == name:
            return service

    exception_message = "You have selected a service that isn't currently live: " + name

    raise Exception(exception_message)
    return None



class Filter:
    def __init__(self):
        self.response = {}


class ServiceDiscoveryDataFilter(Filter):
    def apply(self, response):
        self.response = response

        self.remove_extra_name_field()
        self.correct_dataset_types()
        self.set_empty_parameters_string_to_dict()
        self.wrap_parameters()

        return self.response

    def remove_extra_name_field(self):
        if 'service' not in self.response:
            return

        if 'parameters' not in self.response['service']:
            return

        if 'name' in self.response['service']['parameters']:
            del self.response['service']['parameters']['name']

    def correct_dataset_types(self):
        if 'dataset' not in self.response:
            return

        if 'input' in self.response['dataset']:
            if type(self.response['dataset']['input']) == list:
                value = self.response['dataset']['input'][0]

                self.response['dataset']['input'] = value

        if 'output' in self.response['dataset']:
            if type(self.response['dataset']['output']) == list:
                value = self.response['dataset']['output'][0]

                self.response['dataset']['output'] = value

    def wrap_parameters(self):
        if 'service' not in self.response:
            return

        if 'parameters' not in self.response['service']:
            return

        if 'properties' not in self.response['service']['parameters']:
            value = self.response['service'].pop('parameters')
            self.response['service']['parameters'] = {'properties': value}

    def set_empty_parameters_string_to_dict(self):
        if 'service' not in self.response:
            return

        if 'parameters' not in self.response['service']:
            return

        if self.response['service']['parameters'] == '':
            self.response['service']['parameters'] = {'properties': {}}


class ServiceDiscoveryResponseFilter:
    def apply(self, response):
        service_discovery_data_filter = ServiceDiscoveryDataFilter()

        output = []

        for service_response in response['data']:
            result = service_discovery_data_filter.apply(service_response)
            output.append(result)

        return {'data': output}



from swap.api.commands import (
    ProgressRequestCommand,
    ServiceExecutionRequestCommand,
    ServiceOffersRequestCommand,
    ServiceRequestCommand,
)
from swap.api.hooks import (
    ProgressRequestPostHook,
    ProgressRequestPreHook,
    ServiceExecutionRequestPostHook,
    ServiceExecutionRequestPreHook,
    ServiceOffersRequestPostHook,
    ServiceOffersRequestPreHook,
    ServiceRequestPostHook,
    ServiceRequestPreHook,
)
from swap.api.sequences import CommandSequence
from swap.api.steps import CommandStep


class DefaultCommandSequenceBuilder:
    def __init__(self):
        self.command_sequence = CommandSequence()

    def service_request_step(self):
        return CommandStep(
            command=ServiceRequestCommand(),
            pre_hook=ServiceRequestPreHook(),
            post_hook=ServiceRequestPostHook()
        )

    def service_offers_request_step(self):
        return CommandStep(
            command=ServiceOffersRequestCommand(),
            pre_hook=ServiceOffersRequestPreHook(),
            post_hook=ServiceOffersRequestPostHook()
        )

    def service_execution_request_step(self):
        return CommandStep(
            command=ServiceExecutionRequestCommand(),
            pre_hook=ServiceExecutionRequestPreHook(),
            post_hook=ServiceExecutionRequestPostHook()
        )

    def progress_request_step(self):
        return CommandStep(
            command=ProgressRequestCommand(),
            pre_hook=ProgressRequestPreHook(),
            post_hook=ProgressRequestPostHook()
        )

    def build(self):
        self.command_sequence.add_step(
            'ServiceRequest', self.service_request_step()
        )

        self.command_sequence.add_step(
            'ServiceOffersRequest', self.service_offers_request_step()
        )

        self.command_sequence.add_step(
            'ServiceExecutionRequest', self.service_execution_request_step()
        )

        self.command_sequence.add_step(
            'ProgressRequest', self.progress_request_step()
        )

        return self.command_sequence



class StandardHeader:
    def __init__(
        self, app_message_uuid, request_message_uuid, originating_message_uuid, service_chain_uuid,
        message_timestamp, user_token, user_id, service_chain
    ):
        self.app_message_uuid = app_message_uuid
        self.request_message_uuid = request_message_uuid
        self.originating_message_uuid = originating_message_uuid
        self.service_chain_uuid = service_chain_uuid
        self.message_timestamp = message_timestamp
        self.user_token = user_token
        self.user_id = user_id
        self.service_chain = service_chain


class ServiceProviderHeader:
    def __init__(
        self, service_provider_uuid, service_offer_uuid, service_request_uuid,
        service_provider_signature
    ):
        self.service_provider_uuid = service_provider_uuid
        self.service_offer_uuid = service_offer_uuid
        self.service_request_uuid = service_request_uuid
        self.service_provider_signature = service_provider_signature


class Vendor:
    def __init__(self, service_provider_name, service_provider_uuid, service_provider_signature):
        self.service_provider_name = service_provider_name
        self.service_provider_uuid = service_provider_uuid
        self.service_provider_signature = service_provider_signature


class DatasetDetailsInputComponent:
    def __init__(self, data_token, dataset_uuid, dataset_type, cdf_metadata=''):
        self.data_token = data_token
        self.dataset_uuid = dataset_uuid
        self.dataset_type = dataset_type
        self.cdf_metadata = cdf_metadata


class DatasetDetailsOutputComponent:
    def __init__(self, file_uri, dataset_uuid, dataset_type):
        self.file_uri = file_uri
        self.dataset_uuid = dataset_uuid
        self.dataset_type = dataset_type


class ExecutionDetailsComponent:
    def __init__(self, exec_request_uuid, exec_process_uuid):
        self.exec_request_uuid = exec_request_uuid
        self.exec_process_uuid = exec_process_uuid


class Service:
    def __init__(self, catalogue_uuid, description, parameters, name):
        self.catalogue_uuid = catalogue_uuid
        self.description = description
        self.parameters = parameters
        self.name = name


class Dataset:
    def __init__(self, input, output):
        self.input = input
        self.output = output


class DatasetRequest:
    def __init__(self, data_uri, metadata):
        self.data_uri = data_uri
        self.metadata = metadata


class Icon:
    def __init__(self, base64, url):
        self.base64 = base64
        self.url = url


class ExecutionCompletion:
    def __init__(
        self, standard_header, service_provider_header, execution_details, exec_comp_code,
        exec_comp_text, exec_completion_timestamp, results_dataset
    ):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.execution_details = execution_details
        self.exec_comp_code = exec_comp_code
        self.exec_comp_text = exec_comp_text
        self.exec_completion_timestamp = exec_completion_timestamp
        self.results_dataset = results_dataset


class ExecutionFailure:
    def __init__(
        self, standard_header, service_provider_header, execution_details, exec_fail_code,
        unfulfilled_service_uuid, exec_fail_text
    ):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.execution_details = execution_details
        self.exec_fail_code = exec_fail_code
        self.unfulfilled_service_uuid = unfulfilled_service_uuid
        self.exec_fail_text = exec_fail_text


class ExecutionRequestAccept:
    def __init__(
        self, standard_header, service_provider_header, execution_details,
        exec_start_timestamp
    ):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.execution_details = execution_details
        self.exec_start_timestamp = exec_start_timestamp


class ExecutionRequestCancel:
    def __init__(self, standard_header, service_provider_header, offer_execution_request_to_cancel):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.offer_execution_request_to_cancel = offer_execution_request_to_cancel


class ExecutionRequestDecline:
    def __init__(
        self, standard_header, service_provider_header, execution_details, decline_reason_text,
        exec_decline_timestamp
    ):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.execution_details = execution_details
        self.decline_reason_text = decline_reason_text
        self.exec_decline_timestamp = exec_decline_timestamp


class ExecutionRequest:
    def __init__(
        self, standard_header, service_provider_header, dataset_input_details,
        dataset_output_details, service_offer_description, execution_parameters, service_offer_name
    ):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.dataset_input_details = dataset_input_details
        self.dataset_output_details = dataset_output_details
        self.service_offer_description = service_offer_description
        self.execution_parameters = execution_parameters
        self.service_offer_name = service_offer_name


class OfferExecutionRequest:
    def __init__(
        self, standard_header, service_provider_header, input_data, output_data,
        service_offer_description, execution_parameters, service_offer_name
    ):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.input_data = input_data
        self.output_data = output_data
        self.service_offer_description = service_offer_description
        self.execution_parameters = execution_parameters
        self.service_offer_name = service_offer_name


class ExecutionStatusUpdate:
    def __init__(
        self, standard_header, service_provider_header, execution_details, exec_status_code,
        exec_status_text, exec_update_timestamp
    ):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.execution_details = execution_details
        self.exec_status_code = exec_status_code
        self.exec_status_text = exec_status_text
        self.exec_update_timestamp = exec_update_timestamp


class ServiceOffering:
    def __init__(
        self, standard_header, service_provider_header, input_data,
        output_data, service_offer_description, service_offer_parameters,
        service_offer_name
    ):
        self.standard_header = standard_header
        self.service_provider_header = service_provider_header
        self.input_data = input_data
        self.output_data = output_data
        self.service_offer_description = service_offer_description
        self.service_offer_parameters = service_offer_parameters
        self.service_offer_name = service_offer_name


class ServiceRequest:
    def __init__(
        self, standard_header, input_data, output_data,
        service_catalogue_uuid, permissible_service_provider_uuid
    ):
        self.standard_header = standard_header
        self.input_data = input_data
        self.output_data = output_data
        self.service_catalogue_uuid = service_catalogue_uuid
        self.permissible_service_provider_uuid = permissible_service_provider_uuid


class Message:
    def __init__(self, message_data, message_type, message_raw):
        self.data = message_data
        self.type = message_type
        self.raw = message_raw


class Config:
    def __init__(
        self, project, topic, subscription, catalogue_uuid, dataset_type, icon_base64, icon_url,
        name, service_description, service_provider_name, service_provider_signature,
        service_provider_uuid, parameters
    ):
        self.project = project
        self.topic = topic
        self.subscription = subscription
        self.catalogue_uuid = catalogue_uuid
        self.dataset_type = dataset_type
        self.name = name
        self.service_description = service_description
        self.service_provider_name = service_provider_name
        self.service_provider_signature = service_provider_signature
        self.service_provider_uuid = service_provider_uuid
        self.icon_base64 = icon_base64
        self.icon_url = icon_url
        self.parameters = parameters



from marshmallow import Schema, fields, post_load, EXCLUDE
from swap.common.models import (
    ServiceDiscoveryData,
    ServiceDiscoveryDataset,
    ServiceDiscoveryIcon,
    ServiceDiscoveryParameters,
    ServiceDiscoveryProperty,
    ServiceDiscoveryResponse,
    ServiceDiscoveryService,
    ServiceDiscoveryVendor
)


class ServiceDiscoveryVendorSchema(Schema):
    service_provider_name = fields.Str()
    service_provider_uuid = fields.Str()
    service_provider_signature = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceDiscoveryVendor(**other)


class ServiceDiscoveryPropertySchema(Schema):
    title = fields.Str()
    type = fields.Str()
    options = fields.List(fields.Str())
    description = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceDiscoveryProperty(**other)


class ServiceDiscoveryParametersSchema(Schema):
    properties = fields.Dict(
        keys=fields.Str(), values=fields.Nested(ServiceDiscoveryPropertySchema)
    )

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceDiscoveryParameters(**other)


class ServiceDiscoveryServiceSchema(Schema):
    name = fields.Str()
    catalogue_uuid = fields.Str()
    service_description = fields.Str()
    parameters = fields.Nested(ServiceDiscoveryParametersSchema)

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceDiscoveryService(**other)


class ServiceDiscoveryDatasetSchema(Schema):
    input = fields.Str()
    output = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceDiscoveryDataset(**other)


class ServiceDiscoveryIconSchema(Schema):
    base64 = fields.Str()
    url = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceDiscoveryIcon(**other)


class ServiceDiscoveryDataSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    vendor = fields.Nested(ServiceDiscoveryVendorSchema)
    service = fields.Nested(ServiceDiscoveryServiceSchema)
    dataset = fields.Nested(ServiceDiscoveryDatasetSchema)
    icon = fields.Nested(ServiceDiscoveryIconSchema)

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceDiscoveryData(**other)


class ServiceDiscoveryResponseSchema(Schema):
    data = fields.List(fields.Nested(ServiceDiscoveryDataSchema))

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceDiscoveryResponse(**other)


class ServiceDiscoverySchema(Schema):
    pass



from swap.messaging.clients import MessageSubscriberClient
from swap.messaging.handlers import MessageHandler


class MessageListener:
    def __init__(self, config):
        self.handler = MessageHandler(config=config)
        self.subscriber_client = MessageSubscriberClient(
            project=config.project,
            topic=config.topic,
            subscription=config.subscription
        )

    def callback(self, message):
        message.ack()
        self.handler.handle(message)

    def listen(self):
        future = self.subscriber_client.subscribe(callback=self.callback)

        try:
            future.result()
        except KeyboardInterrupt:
            future.cancel()



import json
from swap.api.clients import UBLClient, IngestionClient
from swap.api.models import ServiceRequest, ServiceRequestChainLink
from swap.api.parsers import ServiceOffersResponseParser, ProgressResponseParser
from swap.api.requests.builders import ServiceExecutionRequestBuilder
from swap.api.schemas import (
    ServiceRequestSchema, ServiceExecutionRequestSchema, IngestionProgressResponseSchema
)
from swap.api.state import ProgressStateMachine
from swap.common.logging import logger
from time import sleep
from swap.api.sequences.step_codes import StepCode


class Command:
    def __init__(self):
        self.ubl_client = UBLClient()


class ServiceRequestCommand(Command):
    def build_service_request(self, input):
        service_request = ServiceRequest(service_chain={})

        for service in input.service_chain.services:
            datasets_as_dicts = []

            for dataset in input.service_chain.datasets:
                datasets_as_dicts.append(dataset.to_dict())

            service_request.add_chain_link(
                ServiceRequestChainLink(
                    datasets=datasets_as_dicts,
                    service=service.name
                )
            )

        return service_request

    def request_body(self, input):
        service_request = self.build_service_request(input)
        schema = ServiceRequestSchema()
        request = schema.dump(service_request)

        logger.debug(f'Built ServiceRequest body: {request}')

        return request

    def submit(self, input):
        request_body = self.request_body(input)
        raw_result = self.ubl_client.post('requests', data=request_body)

        try:
            result_as_json = json.loads(raw_result)
        except:
            raise Exception(f'Failed to load result as json {raw_result}. request_body: {request_body}')

        return result_as_json

    def call(self, input):
        logger.debug('Calling ServiceRequestCommand')
        response = self.submit(input)
        output = input
        output.response = response

        return output


class ServiceOffersRequestCommand(Command):
    def __init__(self):
        self.current_retries = 0
        self.max_retries = 5
        self.sleep_time = 2

        super().__init__()

    def deserialise_response(self, response):
        parser = ServiceOffersResponseParser()
        return parser.parse(response)

    def submit(self, service_chain_uuid):
        if self.current_retries <= self.max_retries:
            url = f'serviceoffers?service_chain_uuid={service_chain_uuid}'
            response = self.ubl_client.get(url)

            logger.debug(f'Raw ServiceOffersResponse: {response}')

            if 'No ServiceRequests received yet' in response:
                logger.debug('Retrying')
                self.current_retries += 1
                sleep(self.sleep_time)

                return self.submit(service_chain_uuid)

            return self.deserialise_response(response)
        else:
            raise Exception(f'Max retries of {self.max_retries} exceeded')

    def call(self, input):
        logger.debug('Calling ServiceOffersRequestCommand')

        service_chain_uuid = input.response['service_chain_uuid']

        logger.info(f'Service chain UUID: {service_chain_uuid}')

        response = self.submit(service_chain_uuid)

        if response == StepCode.NoServiceOffersReceived:
            return StepCode.NoServiceOffersReceived

        output = input
        output.response = response
        return output


class ServiceExecutionRequestCommand(Command):
    def add_parameters(self, request_body_dict, input):
        for chain_link_key, chain_link in request_body_dict['service_chain'].items():
            srvc = next(x for x in input.service_chain.services if x.chain_link == chain_link_key)

            for key, value in srvc.params.configuration.items():
                request_body_dict['service_chain'][chain_link_key]['parameters'][key] = value

        return request_body_dict

    def build_request_body(self, input):
        service_execution_request_builder = ServiceExecutionRequestBuilder()
        service_execution_request = service_execution_request_builder.build(input)

        schema = ServiceExecutionRequestSchema()
        request_body_dict = schema.dump(service_execution_request)

        request_body_dict_with_parameters = self.add_parameters(request_body_dict, input)

        return request_body_dict_with_parameters['service_chain']

    def submit(self, input):
        url = '/serviceexecutions'
        request_body = self.build_request_body(input)
        logger.debug(f'ServiceExecutionRequest body: {request_body}')
        raw_response = self.ubl_client.post(url, data=request_body)

        return json.loads(raw_response)

    def call(self, input):
        logger.debug('Calling ServiceExecutionRequestCommand')
        response = self.submit(input)

        logger.debug(response)

        output = input
        output.response = response
        return output


class ProgressRequestCommand(Command):
    def __init__(self):
        self.parser = ProgressResponseParser()

        super().__init__()

    def submit(self, service_chain_uuid):
        url = f'/progress?service_chain_uuid={service_chain_uuid}'
        response = self.ubl_client.get(url)
        return response

    def service_chain_keys(self, service_chain_uuid):
        response = self.submit(service_chain_uuid)
        parsed_response = self.parser.parse(response)

        return parsed_response.chain_links.keys()

    def call(self, input):
        logger.debug('Calling ProgressRequestCommand')

        service_chain_uuid = input.response['Execution_sent']

        logger.debug(f'Progress update service chain UUID: {service_chain_uuid}')

        output = {}

        for key in self.service_chain_keys(service_chain_uuid):
            state_machine = ProgressStateMachine(
                service_chain_uuid=service_chain_uuid, chain_link=key
            )

            result = state_machine.run()
            output[key] = result

        return output


class IngestionProgressCommand:
    def __init__(self):
        self.client = IngestionClient()

    def submit(self, id):
        result = self.client.get_ingestion_progress(id=id)
        schema = IngestionProgressResponseSchema()
        return schema.load(result)

    def call(self, id):
        result = self.submit(id=id)


# "NONE",
# "QUEUED",
# "IN_PROGRESS",
# "SUCCESS",
# FAILED",
# TIMEOUT"



from marshmallow import Schema, fields, post_load
from swap.messaging.models import (
    Dataset,
    DatasetDetailsInputComponent,
    DatasetDetailsOutputComponent,
    ExecutionCompletion,
    ExecutionDetailsComponent,
    ExecutionFailure,
    ExecutionRequest,
    ExecutionRequestAccept,
    ExecutionRequestCancel,
    ExecutionRequestDecline,
    ExecutionStatusUpdate,
    Icon,
    OfferExecutionRequest,
    Service,
    ServiceOffering,
    ServiceProviderHeader,
    ServiceRequest,
    StandardHeader,
    Vendor,
    DatasetRequest
)


class StandardHeaderSchema(Schema):
    app_message_uuid = fields.Str()
    request_message_uuid = fields.Str()
    originating_message_uuid = fields.Str()
    service_chain_uuid = fields.Str()
    message_timestamp = fields.Str()
    user_token = fields.Str()
    user_id = fields.Str()
    service_chain = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return StandardHeader(**other)


class ServiceProviderHeaderSchema(Schema):
    service_provider_uuid = fields.Str()
    service_offer_uuid = fields.Str()
    service_request_uuid = fields.Str()
    service_provider_signature = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceProviderHeader(**other)


class VendorSchema(Schema):
    service_provider_name = fields.Str()
    service_provider_uuid = fields.Str()
    service_provider_signature = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return Vendor(**other)


class DatasetRequestSchema(Schema):
    data_uri = fields.Str(allow_none=True)
    metadata = fields.Str(allow_none=True)

    @post_load
    def deserialise(self, other, **kwargs):
        return DatasetRequest(**other)


class DatasetDetailsInputComponentSchema(Schema):
    cdf_metadata = fields.Str(allow_none=True)
    data_token = fields.Str(allow_none=True)
    dataset_type = fields.Str(allow_none=True)
    dataset_uuid = fields.Nested(DatasetRequestSchema)

    @post_load
    def deserialise(self, other, **kwargs):
        return DatasetDetailsInputComponent(**other)


class DatasetDetailsOutputComponentSchema(Schema):
    file_uri = fields.Str()
    dataset_uuid = fields.Str()
    dataset_type = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return DatasetDetailsOutputComponent(**other)


class ExecutionDetailsComponentSchema(Schema):
    exec_request_uuid = fields.Str()
    exec_process_uuid = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionDetailsComponent(**other)


class ServiceSchema(Schema):
    catalogue_uuid = fields.Str()
    description = fields.Str()
    parameters = fields.Str()
    name = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return Service(**other)


class DatasetSchema(Schema):
    input = fields.Str()
    output = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return Dataset(**other)


class IconSchema(Schema):
    base64 = fields.Str()
    url = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return Icon(**other)


class ExecutionCompletionSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    execution_details = fields.Nested(ExecutionDetailsComponentSchema)
    exec_comp_code = fields.Str()
    exec_comp_text = fields.Str()
    exec_completion_timestamp = fields.Str()
    results_dataset = fields.Nested(DatasetDetailsOutputComponentSchema)

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionCompletion(**other)


class ExecutionFailureSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    execution_details = fields.Nested(ExecutionDetailsComponentSchema)
    exec_fail_code = fields.Str()
    unfulfilled_service_uuid = fields.Str()
    exec_fail_text = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionFailure(**other)


class ExecutionRequestAcceptSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    execution_details = fields.Nested(ExecutionDetailsComponentSchema)
    exec_start_timestamp = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionRequestAccept(**other)


class ExecutionRequestCancelSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    offer_execution_request_to_cancel = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionRequestCancel(**other)


class ExecutionRequestDeclineSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    execution_details = fields.Nested(ExecutionDetailsComponentSchema)
    decline_reason_text = fields.Str()
    exec_decline_timestamp = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionRequestDecline(**other)


class ExecutionRequestSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    dataset_input_details = fields.List(fields.Nested(DatasetDetailsInputComponentSchema))
    dataset_output_details = fields.Nested(DatasetDetailsOutputComponentSchema)
    service_offer_description = fields.Str()
    execution_parameters = fields.Str()
    service_offer_name = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionRequest(**other)


class OfferExecutionRequestSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    input_data = fields.List(fields.Nested(DatasetDetailsInputComponentSchema))
    output_data = fields.Nested(DatasetDetailsOutputComponentSchema)
    service_offer_description = fields.Str()
    execution_parameters = fields.Str()
    service_offer_name = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return OfferExecutionRequest(**other)


class ExecutionStatusUpdateSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    execution_details = fields.Nested(ExecutionDetailsComponentSchema)
    exec_status_code = fields.Str()
    exec_status_text = fields.Str()
    exec_update_timestamp = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ExecutionStatusUpdate(**other)


class ServiceOfferingSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    service_provider_header = fields.Nested(ServiceProviderHeaderSchema)
    input_data = fields.List(fields.Nested(DatasetDetailsInputComponentSchema))
    output_data = fields.Nested(DatasetDetailsOutputComponentSchema)
    service_offer_description = fields.Str()
    service_offer_parameters = fields.Str()
    service_offer_name = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceOffering(**other)


class ServiceRequestSchema(Schema):
    standard_header = fields.Nested(StandardHeaderSchema)
    input_data = fields.List(fields.Nested(DatasetDetailsInputComponentSchema))
    output_data = fields.Nested(DatasetDetailsOutputComponentSchema)
    service_catalogue_uuid = fields.Str()
    permissible_service_provider_uuid = fields.Str()

    @post_load
    def deserialise(self, other, **kwargs):
        return ServiceRequest(**other)



from swap.common.logging import logger
from swap.api.sequences.step_codes import StepCode

class ServiceRequestPreHook:
    def call(self, input):
        return input


class ServiceRequestPostHook:
    def call(self, input):
        return input


class ServiceOffersRequestPreHook:
    def call(self, input):
        return input


class ServiceOffersRequestPostHook:
    def select_first_service_offers(self, input):
        chain_links = input.response.chain_links

        response = {}

        for chain_link_key, chain_link in chain_links.items():
            if not chain_link.service_offers.items():
                logger.info(f'No service offers received yet for service: {chain_link.service_catalogue_uuid}')
                return StepCode.NoServiceOffersReceived
            service_offers = list(chain_link.service_offers.values())
            response[chain_link_key] = service_offers[0]

        output = input
        output.response = response

        return output

    def call(self, input):
        return self.select_first_service_offers(input)


class ServiceExecutionRequestPreHook:
    def call(self, input):
        return input


class ServiceExecutionRequestPostHook:
    def call(self, input):
        return input


class ProgressRequestPreHook:
    def call(self, input):
        return input


class ProgressRequestPostHook:
    def call(self, input):
        return input



from swap.messaging.registries import DefaultRegistry


class MessageHandler:
    def __init__(self, config, registry=DefaultRegistry):
        self.config = config
        self.registry = registry(config=config)

    def register_callback(self, message_type, callback):
        self.registry.register_callback(message_type=message_type, callback=callback)

    def action(self, message_type):
        return self.registry.action(message_type)

    def callback(self, message_type):
        return self.registry.callback(message_type)

    def handle(self, message):
        message_type = message.attributes.get('message_type')
        action = self.action(message_type=message_type)
        callback = self.callback(message_type=message_type)

        action.call(message_type=message_type, message=message, callback=callback)



from environs import Env


env = Env()


class Settings:
    BUCKET = env.str('BUCKET', 'bucket-name')
    DAS_URL = env.str('DAS_URL', 'https://www.akerbp.com')
    LOG_LEVEL = env.str('LOG_LEVEL', 'INFO')
    SERVICE_DISCOVERY_URL = env.str('SERVICE_DISCOVERY_URL', 'https://www.akerbp.com')
    SWAP_API_TOKEN = env.str('SWAP_API_TOKEN')
    UBL_URL = env.str('UBL_URL', 'https://www.akerbp.com')



import setuptools

# with open("README.md", "r") as fh:
#     long_description = fh.read()

setuptools.setup(
    name='swap-python-sdk',
    version='0.1.0',
    author='',
    author_email='',
    description='SWAP Python SDK',
    # long_description=long_description,
    long_description_content_type='text/markdown',
    url='',
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[
        'google-cloud-pubsub==1.6.0',
        'jsonschema==3.2.0',
        'marshmallow==3.3.0',
        'python-dotenv==0.10.5',
        'requests==2.22.0',
        'environs==8.0.0'
    ],
    package_data={'swap': ['common/static/schemas/*']},
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ]
)
